/*
 * Copyright 2002 Henri Sivonen.
 *
 * The contents of this file are subject to the Artistic License.
 * You may not use this file except in compliance with the Artistic
 * License (which you should have received along with this file).
 */

package dualroids;

import java.awt.*;
import java.awt.geom.*;

/**
 * The representation of a bullet.
 *
 * <h3>A Note About Thread Safety</h3>
 * <p>The implementation of <code>Bullet</code> is thread-safe
 * in the sense that the one thread may call the setters
 * while another calls the getters. However, there may
 * <strong>not</strong> be multiple threads calling
 * <code>getTransformedShape</code>. It is designed
 * to be used from a single thread only.</p>
 * <p>Also, that thread should not retain references to
 * objects returned by <code>getTransformedShape</code>
 * across the method calls. When the method is called
 * the object it returned previously will no longer be valid.
 * </p>
 * 
 * @author Henri Sivonen, hsivonen@iki.fi
 */
public final class Bullet extends SpatialObject {

    /**
     * Constant for the bullet velocity.
     */
    private static double BULLET_VELOCITY = 0.1;

    /**
     * Constant for the length of a bullet divided by 2.
     */
    private static double BULLET_HALF_LENGTH = 2.5;

    /**
     * The lifetime of a bullet in milliseconds.
     */
    private static long BULLET_TTL = 3000L;

    /**
     * The <code>Point2D</code> instance used for returning
     * the translated location of the bullet
     */
    private Point2D translatableLocation;

    /**
     * The x difference of the between the center of bullet the
     * end of the bullet line
     */        
    private double lineDeltaX;

    /**
     * The y difference of the between the center of bullet the
     * end of the bullet line
     */
    private double lineDeltaY;

    /**
     * The <code>Line2D</code> instance used for returning
     * the translated <code>Shape</code> representation of the bullet
     */
    private Line2D translatableLine;

    /**
     * The moment in time when this bullet expires
     */
    private long expires;
    
    /**
     * Contructs a new bullet. The fields aren't initialized
     * to any meaningful values. The values are supplied
     * using the <code>fire</code> method.
     */
    public Bullet() {
        super(0.0, 0.0, 0.0, 0.0, -1L);
        this.translatableLocation = new Point2D.Double();
        this.lineDeltaX = 0.0;
        this.lineDeltaY = 0.0;
        this.translatableLine = new Line2D.Double();
        this.expires = -1L;
    }

    /**
     * Returns the location of the bullet at <code>time</code>.
     * There should be one thread at maximum per arena calling
     * this method. The returned Point2D shouldn't be retained
     * as it will become invalid when this method is called again.
     *
     * @param time the moment in game (server) time for which the
     *             location is requested
     * @return a <code>Point2D</code>
     */
    public final Point2D getLocation(long time) {
        if(time > this.expires) {
            this.expires = -1L; // this is an ugly hack for readyToBeFired
            return null;
        } else {
            synchronized(this) {
                double deltaT = (double)(time - this.timestamp);
                this.translatableLocation.setLocation(moveToRangeX(this.locationX
                                                                   + this.velocityX
                                                                   * deltaT),
                                                      moveToRangeY(this.locationY
                                                                   + this.velocityY
                                                                   * deltaT));
            }
            return this.translatableLocation;
        }
    }

    /**
     * Fires this bullet. Used only on the client that controls
     * the ship that owns this bullet.
     *
     * @param locationX the x coordinate of the ship
     * @param locationY the y coordinate of the ship
     * @param shipVelocityX the x component of the velocity of the ship
     * @param shipVelocityY the y component of the velocity of the ship
     * @param angle the angle of the ship
     * @param timestamp the moment of firing
     */
    public final void fire(double locationX,
                           double locationY,
                           double shipVelocityX,
                           double shipVelocityY,
                           double angle,
                           long timestamp)
    {
        synchronized(this) {
            double sinAngle = Math.sin(angle);
            double cosAngle = Math.cos(angle);
            this.locationX = locationX;
            this.locationY = locationY;
            this.velocityX = shipVelocityX + BULLET_VELOCITY * cosAngle;
            this.velocityY = shipVelocityY + BULLET_VELOCITY * sinAngle;
            this.lineDeltaX = BULLET_HALF_LENGTH * cosAngle;
            this.lineDeltaY = BULLET_HALF_LENGTH * sinAngle;
            this.timestamp = timestamp;
            this.expires = timestamp + BULLET_TTL;
        }
    }

    /**
     * Returns a properly transformed <code>Shape</code> representing
     * the spatial object at <code>time</code>
     *
     * @param time the moment in game (server) time for which the
     *             representation is requested
     * @return the <code>Shape</code> representation at <code>time</code>
     *         or null if the object isn't visible
     */
    public final Shape getTransformedShape(long time) {
        if(time > this.expires) {
            this.expires = -1L; // this is an ugly hack for readyToBeFired
            return null;
        } else {
            double deltaT, x, y;
            synchronized(this) {
                deltaT = (double)(time - this.timestamp);
                x = moveToRangeX(this.locationX
                                 + this.velocityX
                                 * deltaT);
                y = moveToRangeY(this.locationY
                                 + this.velocityY
                                 * deltaT);
            }
            this.translatableLine.setLine(x + this.lineDeltaX,
                                          y + this.lineDeltaY,
                                          x - this.lineDeltaX,
                                          y - this.lineDeltaY);
            return this.translatableLine;
        }
    }

    /**
     * Marks the bullet non-flying (not visible)
     */
    public final void invalidate() {
        this.expires = -1L;
    }

    /**
     * Returns a boolean indicating whether the bullet is ready to be fired.
     *
     * @return a boolean indicating whether the bullet is ready to be fired
     */
    public final boolean readyToBeFired() {
        if(this.expires == -1L) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * Writes the state of the bullet to a <code>ClientModelConnector</code>
     *
     * @param thisIndex the index of this bullet
     * @param connector the connector to which the state is to be written
     */
    public final void writeToConnector(int thisIndex,
                                       ClientModelConnector connector)
    {
        try { 
            synchronized(this) {
                connector.writeBulletFired(thisIndex,
                                           this.locationX,
                                           this.locationY,
                                           this.velocityX,
                                           this.velocityY,
                                           this.lineDeltaX,
                                           this.lineDeltaY,
                                           this.timestamp,
                                           this.expires);
            }
            connector.flush();
        } catch(java.io.IOException ioe) {
            Runner.fatalError("Error in network connection!");
        }
    }

    /**
     * Initializes the state of the bullet from external state data.
     *
     * @param locationX the x coordinate at which the bullet was fired
     * @param locationY the y coordinate at which the bullet was fired
     * @param velocityX the x component of the velocity of the bullet
     * @param velocityY the y component of the velocity of the bullet
     * @param timestamp the moment at which the bullet was fired
     * @param expires the moment in time at which the bullet expires
     */
    public final void initFromRemote(double locationX,
                                     double locationY,
                                     double velocityX,
                                     double velocityY,
                                     double lineDeltaX,
                                     double lineDeltaY,
                                     long timestamp,
                                     long expires)
    {
        synchronized(this) {
            this.locationX = locationX;
            this.locationY = locationY;
            this.velocityX = velocityX;
            this.velocityY = velocityY;
            this.lineDeltaX = lineDeltaX;
            this.lineDeltaY = lineDeltaY;
            this.timestamp = timestamp;
            this.expires = expires;
        }
    }
}
