/*
 * Copyright 2002 Yrj Kari-Koskinen.
 *
 * The contents of this file are subject to the Artistic License.
 * You may not use this file except in compliance with the Artistic
 * License (which you should have received along with this file).
 */

package dualroids;

import java.io.IOException;

/**
 * The helper class for running Client threads.
 *
 * @author Yrj Kari-Koskinen, yrjo.kari-koskinen@hut.fi
 */
public class Client implements Runnable {

    /** the host name to connect the client to */
    private String hostname;

    /** whether the client is local to the server or not */
    private boolean local;

    /** the main game frame */
    private GameFrame gameFrame;

    /** the player's name of this client */
    private String playerName;


    /**
     * Saves the <code>hostname</code>, <code>local</code> and 
     * <code>playerName</code> fields.
     *
     * @param hostname the host name to connect the client to
     * @param local whether the client is local to the server or not
     * @param playerName the player's name of this client
     */
    public Client(String hostname, boolean local, String playerName) {
        this.hostname = hostname;
        this.local = local;
        this.playerName = playerName;
    }


    /**
     * JVM calls this method when the client thread is started with the
     * <code>start()</code> method call. First the network connections are
     * initialized by creating a client model connector and then the main
     * game window is initialized.
     */
    public void run() {
        ClientModelConnector cmc = null;
        try {
            // create new cmc, does also the pinging if necessary
            cmc = ClientModelConnector.
                createClientModelConnector(this.hostname, 
                                           this.local,
                                           this.playerName);
        } catch(IOException ioe) {
            Runner.getInstance().dispose();
            Runner.fatalError("Error in connecting to " +
                              "Dualroids server!\n" +
                              ioe.getMessage());
        }

        //initialize main game frame
        this.gameFrame = GameFrame.getInstance(cmc);
	
	if (this.local) {
	    new Thread(cmc, "local cmc").start();
	} else {
	    new Thread(cmc, "non-local cmc").start();
	}

        // if client is not local, show the game frame immediately
        // otherwise it will not be made visible until it's suitable
        if (!this.local) {
            this.setVisible();
        }
    }


    /**
     * This method gets rid of the setup screen windows and shows the 
     * main game window.
     */
    public void setVisible() {
        // get rid of the setup screen windows
        Runner.getInstance().dispose();

        // show the game frame window
        this.gameFrame.setVisible(true);
    }
}
