/*
 * Copyright 2002 Henri Sivonen.
 *
 * The contents of this file are subject to the Artistic License.
 * You may not use this file except in compliance with the Artistic
 * License (which you should have received along with this file).
 */

package dualroids;

import java.io.IOException;
import java.awt.*;
import java.awt.geom.*;

/**
 * The Dualroids game server.
 *
 * @author Henri Sivonen, hsivonen@iki.fi
 */
public final class Server implements Runnable {

    /**
     * Constant for the minimal interval (in milliseconds) between
     * the starts of two server model update iterations.
     */
    public static final long MIN_ITERATION_INTERVAL = 16L;

    /**
     * The delay between the beginning of level generation
     * and the start of the level.
     */
    private static final long LEVEL_START_DELAY = 2000L;

    /**
     * The duration of the ship invulnerability period when
     * a player gets a new ship.
     */
    public static final long INVULNERABILITY_PERIOD = 3000L;

    /**
     * The delay before a new ships appears when a ship has
     * exploded.
     */
    public static final long NEW_SHIP_DELAY = 2000L;

    /**
     * The number of spare ships the player has initially.
     */
    public static final int INITIAL_SHIPS = 5;

    /**
     * The interval in scores when player should get an additional
     * spare ship
     */
    private static final int SCORE_1UP = 500;
    
    /**
     * The server's copy of the arena
     */
    private Arena arena;

    /**
     * An array holding all the connectors through which the server
     * sends data to the clients.
     */
    private ServerModelConnector[] connectors;

    /**
     * The level number counter
     */
    private int level;

    /**
     * A counter for controlling the asteroid updating iteration.
     */
    private int asteroidIndex;

    /**
     * The score table.
     */
    private int[] scores;

    /**
     * The table of remaining ships.
     */
    private int[] shipsLeft;

    /**
     * The name of the player who is running the server
     */
    private String playerName;

    /**
     * The method that starts the server thread.
     */
    public void run() {
        try {
            this.connectors = 
                ServerModelConnector.
                createServerModelConnectors(this.playerName);
        } catch(IOException ioe) {
            Runner.getInstance().dispose();
            Runner.fatalError("Failed to create Dualroids server!\n" +
                              ioe.getMessage());
        }
        try {
            for(;;) {
                long levelStarts = System.currentTimeMillis()
                                   + LEVEL_START_DELAY;
                this.arena = new Arena(this.level, levelStarts);
                this.sendArena();
                this.flush();
                long timeToSleep = levelStarts - System.currentTimeMillis();
                if(timeToSleep > 0L) {
                    try {
                        Thread.sleep(timeToSleep);
                    } catch(InterruptedException ie) {
                        // carry on
                    }
                }
                this.runLevel();
                this.level++;
            }
        } catch(GameOverException goe) {
            try {
                this.connectors[0].writeGameOver();
            } catch(IOException ioe) {
                // we don't care if writeGameOver() throws an exception
            }
        }
    }

    /**
     * Contructs a new <code>Server</code>.
     *
     * @param playerName the name of the player running the server
     */
    public Server(String playerName) {
        this.playerName = playerName;
        this.arena = null;
        this.level = 1;
        this.asteroidIndex = 0;
        this.scores = new int[2];
        this.shipsLeft = new int[2];
        this.shipsLeft[0] = this.shipsLeft[1] = Server.INITIAL_SHIPS;
        this.scores[0] = this.scores[1] = 0;
    }
    
    /**
     * The method for running a level
     */
    private final void runLevel() throws GameOverException {
        long timeOfIteration, timeToSleep;
        boolean asteroidsLeft = true;
        for(;;) {
            timeOfIteration = System.currentTimeMillis();

            this.checkCollisions(timeOfIteration);
            asteroidsLeft = this.updateAsteroids(timeOfIteration);
            this.flush();
            if(!asteroidsLeft) {
                return;
            }
            
            timeToSleep = timeOfIteration
                + MIN_ITERATION_INTERVAL
                - System.currentTimeMillis();
            /*
             * Making sure the thread at least yields even if there is
             * no time to sleep.
             */
            if(timeToSleep < 1L) {
                Thread.yield();
            } else {
                try {
                    Thread.sleep(timeToSleep);
                } catch(InterruptedException ie) {
                    // carry on
                }
            }

        }
    }

    /**
     * Flushes both connectors and thus causes the buffered data in network
     * streams to flush too.
     */
    private final void flush() {
        for(int i = 0; i < this.connectors.length; i++) {
            try {
                this.connectors[i].flush();
            } catch(IOException ioe) {
                Runner.fatalError("Error in network connection!");
            }
        }
    }

    /**
     * Checks collisions in the arena (ships or bullets hitting asteroids).
     *
     * @param time the moment in time for which the check is to be performed
     */
    private final void checkCollisions(long time) throws GameOverException {
        Asteroid[] asteroids = this.arena.getAsteroids();
        SpaceShip[] ships = this.arena.getSpaceShips();
        Bullet[] bullets;
        Asteroid asteroid;
        Shape asteroidGeometry;
        Point2D[] vertices;
        Point2D bulletLocation;
        SpaceShip ship;
        for(int i = 0; i < asteroids.length; i++) {
            asteroid = asteroids[i];
            asteroidGeometry = asteroid.getTransformedShape(time);
            if(asteroidGeometry != null) {
                for(int j = 0; j < ships.length; j++) {
                    ship = ships[j];
                    vertices = ship.getVertices(time);
                    if(vertices != null) {
                        for(int k = 0; k < vertices.length; k++) {
                            if(asteroidGeometry.contains(vertices[k])) {
                                this.blowUpShip(j, time + NEW_SHIP_DELAY);
                                break;
                            }
                        }
                    }
                    bullets = ship.getBullets();
                    for(int k = 0; k < bullets.length; k++) {
                        bulletLocation = bullets[k].getLocation(time);
                        if(bulletLocation != null &&
                           asteroidGeometry.contains(bulletLocation))
                        {
                            bullets[k].invalidate();
                            this.invalidateBullet(j, k);
                            this.registerAsteroidHit(j, i, time);
                        }
                    }
                }
            }
        }
    }

    /**
     * Updates the fields of one asteroid at time and checks whether
     * there are asteroids left.
     *
     * @param time the moment in time to be used as the new timestamp
     * @return <code>true</code> if there are asteroids left and
     *         <code>false</code> otherwise
     */
    public final boolean updateAsteroids(long time) {
        Asteroid[] asteroids = this.arena.getAsteroids();
        int i = this.asteroidIndex;
        boolean wrappedAround = false;
        do {
            i++;
            i %= asteroids.length;
            if(i == this.asteroidIndex) {
                if(wrappedAround) {
                    return false;
                } else {
                    wrappedAround = true;
                }
            }
        } while(!asteroids[i].isVisible());
        this.asteroidIndex = i;
        asteroids[i].updateLocationAndAngle(time);
        for(int j = 0; j < this.connectors.length; j++) {
            asteroids[i].writeToConnector(i, this.connectors[j]);
        }
        return true;
    }
    
    /**
     * Blows up a ship
     *
     * @param shipIndex the index of the ship to blow up
     */
    private final void blowUpShip(int shipIndex, long newAppears)
        throws GameOverException
    {
        this.shipsLeft[shipIndex]--;
        this.arena.getSpaceShips()[shipIndex].explode(newAppears
                                                      + INVULNERABILITY_PERIOD);
        for(int i = 0; i < this.connectors.length; i++) {
            try {
                this.connectors[i].writeBlowUpShip(shipIndex,
                                                   newAppears);
                this.connectors[i].writeShipsLeft(shipIndex,
                                                  this.shipsLeft[shipIndex]);
            } catch(IOException ioe) {
                Runner.fatalError("Error in network connection!");
            }
        }
        if(this.shipsLeft[shipIndex] == -1) {
            throw new GameOverException();
        }
    }

    /**
     * Marks a bullet non-flying (not visible).
     *
     * @param shipIndex the index of the ship owning the bullet
     * @param bulletIndex the index of the bullet
     */
    private final void invalidateBullet(int shipIndex, int bulletIndex) {
        for(int i = 0; i < this.connectors.length; i++) {
            try {
                this.connectors[i].writeInvalidateBullet(shipIndex, 
                                                         bulletIndex);
            } catch(IOException ioe) {
                Runner.fatalError("Error in network connection!");
            }
        }
    }

    /**
     * Register a bullet hit to an asteroid. The hit may cause the asteroid
     * to explode and points to be awarded.
     *
     * @param shipIndex the index of the ship that fired the bullet
     * @param asteroidIndex the index of the asteroid
     */
    private final void registerAsteroidHit(int shipIndex,
                                           int asteroidIndex,
                                           long time)
    {
        try {
            Asteroid a = this.arena.getAsteroids()[asteroidIndex];
            if(a.registerHit()) {
                int oldScore = this.scores[shipIndex];
                int newScore = oldScore + a.getPoints();
                this.scores[shipIndex] = newScore;

                // 1 UP every SCORE_1UP
                if ( (oldScore % SCORE_1UP) > (newScore % SCORE_1UP) ) {
                    this.shipsLeft[shipIndex]++;
                    for(int i = 0; i < this.connectors.length; i++) {
                        this.connectors[i].writeShipsLeft(shipIndex,
                                                          this.shipsLeft[shipIndex]);
                    }
                }

                double locationX = a.getLocationX(time);
                double locationY = a.getLocationY(time);
                Asteroid.blowUpAsteroid(this.arena.getAsteroids(),
                                        asteroidIndex,
                                        locationX,
                                        locationY,
                                        time,
                                        true);
                for(int i = 0; i < this.connectors.length; i++) {
                    this.connectors[i].writeBlowUpAsteroid(asteroidIndex,
                                                           locationX,
                                                           locationY,
                                                           time);
                    this.connectors[i].writeScore(shipIndex,
                                                  this.scores[shipIndex]);
                }
            }
        } catch(IOException ioe) {
            Runner.fatalError("Error in network connection!");
        }
    }

    /**
     * Sends the arena to the clients.
     */
    private final void sendArena() {
        for(int i = 0; i < this.connectors.length; i++) {
            try {
                this.connectors[i].writeArena(this.arena);
            } catch(IOException ioe) {
                Runner.fatalError("Error in network connection!");
            }
        }
    }
}
